#ifndef _Preprocessor_H__
#define _Preprocessor_H__

#include <string>
using namespace std;
#include "Containers.h"
#include "Mesh.h"
#include "KdTree.h"

#include <QObject>

namespace GtpVisibilityPreprocessor {

class RenderSimulator;
class SceneGraph;
class Exporter;
class ViewCellsManager;
class BspTree;
class VspKdTree;
class VspBspTree;
class RenderSimulator;
struct VssRayContainer;

class GlRendererBuffer;

/** Namespace for the external visibility preprocessor

    This namespace includes all classes which are created by the VUT (Vienna University
    of Technology) for the External Visibility Preprocessor of the GTP (GameTools Project)
    (www.gametools.org).
*/
  
  /** Main class of the visibility preprocessor. Responsible for loading and
      saving of the input and output files. Initiates construction of the kD-tree,
      viewcell loading/generation and the visibility computation itself.
  */
class Preprocessor : public QObject {
  Q_OBJECT

public:
	/** Default constructor initialising e.g., KD tree and BSP tree.
	*/
	Preprocessor();

   virtual ~Preprocessor();

  /** Load the input scene. 
      @param filename file to load
	@return true on success
  */
  virtual bool LoadScene(const string filename);
  
  /** Export all preprocessed data in a XML format understandable by the
      PreprocessingInterface of the GtpVisibilityPreprocessor Module. The file can be compressed depending
      on the environement settings.
      @return true on successful export
  */
  virtual bool ExportPreprocessedData(const string filename);
  
  /** Build the KdTree of currently loaded occluders/occludees/viewcells. The construction
      is driven by the environment settings, which also sais which of the three types of
      entities should be used to drive the heuristical construction (only occluders by default)
  */
  virtual bool BuildKdTree();

  /** Compute visibility method. This method has to be reimplemented by the actual
      Preprocessor implementation (e.g. SamplingPreprocessor, ExactPreprocessor,
      GlobalSamplingpreprocessor)
    */
  virtual bool ComputeVisibility() = 0;
  
  /** View cells are either loaded or prepared for generation, according to the chosen environment
      object. Important evironment options are, e.g, the view cell type.
      Should be done after scene loading (i.e., some options are based on scene type).
  */
  bool PrepareViewCells();
  
  bool
  Export( const string filename,
	  const bool scene,
	  const bool kdtree,
	  const bool bsptree
	  );
  
  
  virtual void KdTreeStatistics(ostream &s);
  virtual void BspTreeStatistics(ostream &s);

  /** Loads samples from file.
	  @param samples returns the stored sample rays
	  @param objects needed to associate the objects ids
	  @returns true if samples were loaded successfully
  */
  bool LoadSamples(VssRayContainer &samples,
				   ObjectContainer &objects) const;

  /** Exports samples to file.
	  @returns true if samples were written successfully
  */
  bool ExportSamples(const VssRayContainer &samples) const;

  /** Get Sample rays of particular type, returns false if this
	  type of rays is not supported by the preprocessor
  */
  enum {
	OBJECT_BASED_DISTRIBUTION,
	DIRECTION_BASED_DISTRIBUTION,
	DIRECTION_BOX_BASED_DISTRIBUTION,
	SPATIAL_BOX_BASED_DISTRIBUTION,
	RSS_BASED_DISTRIBUTION,
	RSS_SILHOUETTE_BASED_DISTRIBUTION,
	VSS_BASED_DISTRIBUTION,
	OBJECT_DIRECTION_BASED_DISTRIBUTION
  };
  
  virtual bool
  GenerateRays(
			   const int number,
			   const int raysType,
			   SimpleRayContainer &rays
			   );
  
  virtual void CastRays(SimpleRayContainer &rays,
	  VssRayContainer &vssRays) {};

  ViewCellsManager *CreateViewCellsManager(const char *name);

  /// scene graph loaded from file
  SceneGraph *mSceneGraph;

  /// raw array of objects
  ObjectContainer mObjects;
  
  /// kD-tree organizing the scene graph (occluders + occludees) + viewcells
  KdTree *mKdTree;
  
  /// View space partition bsp tree
  VspBspTree *mVspBspTree;
  /// list of all loaded occluders
  ObjectContainer mOccluders;
  /// list of all loaded occludees
  ObjectContainer mOccludees;
   
  /// BSP tree representing the viewcells
  BspTree *mBspTree;

  ViewCellsManager *mViewCellsManager;

  /** Kd tree inducing a coarse partition of view space that are the building
	  blocks for view cells.
  */
  VspKdTree *mVspKdTree;

  bool mUseGlRenderer;
  bool mUseGlDebugger;

  bool mLoadViewCells;
  
  bool mDetectEmptyViewSpace;

  bool mQuitOnFinish;
  bool mLoadPolygonsAsMeshes;
  bool mComputeVisibility;

  GlRendererBuffer *GetRenderer() { return renderer;}
  
protected:

  /////////////////////////

  /// samples used for construction of the BSP view cells tree.
  int mBspConstructionSamples;
   /// samples used for construction of the VSP KD tree.
  int mVspKdConstructionSamples;
  /** Simulates rendering of the scene.
  */
  RenderSimulator *mRenderSimulator;

  GlRendererBuffer *renderer;
  
  signals:
  void EvalPvsStat();

};



extern Preprocessor *preprocessor;

}

#endif
